interface DerivationNode {
  sequent: string;
  rule?: string;
  ruleInput?: string;
  axiom?: string;
  lemmaTaskId?: number;
  children?: DerivationNode[];
  path?: string;
}

interface InferenceRule {
  top: string;
  bottom: string;
  symbol: string;
  symbolHtml: string;
}

class InferenceRules {
  static symbolToHtml(symbol: string): string {
    return symbol.replace(/\//g, '<span class="stroke"></span>').replace(/_(.)/g, '<sub>$1</sub>');
  }
}

class DerivationNodes {
  static setPaths(node: DerivationNode, path: string) {
    node.path = path;
    if (node.children)
      node.children.forEach((child, i) => {
        DerivationNodes.setPaths(child, path + i);
      });
  }
}

interface DerivationDirectiveScope extends ng.IScope {
  derivation: DerivationNode;
  selectedNode: DerivationNode;
  canInteract: boolean;
  getLemmaLabel(args: {lemmaTaskId: number}): string;
}

interface RenderedDerivationNode extends DerivationNode {
  children?: RenderedDerivationNode[];
  element: JQuery;
  coord: number;
  parent?: RenderedDerivationNode;
  completed: boolean;
  phantom?: boolean;
}

DerivationDirective.$inject = ['$timeout'];
function DerivationDirective($timeout: ng.ITimeoutService): ng.IDirective {
  function renderNode(node: DerivationNode, selectedNode: DerivationNode, getLemmaLabel: (taskId: number) => string): RenderedDerivationNode {
    var el = $('<div class="node"></div>');

    var renderedNode: RenderedDerivationNode = {
      sequent: node.sequent,
      rule: node.rule,
      children: [],
      element: el,
      coord: 0,
      completed: !!node.axiom || !!node.lemmaTaskId
    };

    if (node.children && node.children.length > 0) {
      var ch = $('<div class="children"></div>').appendTo(el);
      node.children.forEach(childNode => {
        var renderedChild = renderNode(childNode, selectedNode, getLemmaLabel);
        renderedChild.parent = renderedNode;
        renderedNode.children.push(renderedChild);
        ch.append(renderedChild.element);
      });
      renderedNode.completed = renderedNode.children.every(rc => rc.completed);
    }
    
    if (node.rule || node.axiom || node.lemmaTaskId) {
      var symbol = node.lemmaTaskId ? getLemmaLabel(node.lemmaTaskId) : InferenceRules.symbolToHtml(node.axiom || node.rule);
      el.append($('<div class="rule"></div>')
          .toggleClass('axiom', !!node.axiom || !!node.lemmaTaskId)
          .append(symbol
              ? $('<span class="symbol"></span>')
                  .append('(', $('<span class="sequent"></span>').toggleClass('sequent', !!node.axiom || !!node.rule).html(symbol), ')')
              : null));
    }
    
    el.append($('<div class="sequent"></div>').text(node.sequent).toggleClass('active', node == selectedNode));
    
    el.data('node', node).toggleClass('completed', renderedNode.completed);

    return renderedNode;
  }

  function layOutTree(rootNode: RenderedDerivationNode) {
    var gap = 60;
    var layers: RenderedDerivationNode[][] = [[rootNode]];
    var leftMost = Number.POSITIVE_INFINITY;
    var rightMost = Number.NEGATIVE_INFINITY;

    for (var l = 0; l < layers.length; ++l) {
      for (var ll = l; ll > 0; --ll) {
        var nodes = layers[ll];
        
        // Move overlapping nodes
        var left = Number.NEGATIVE_INFINITY;
        nodes.forEach((node, i) => {
          var diff = node.coord - left;
          if (diff < 0) {
            for (var j = i; j < nodes.length; ++j)
              nodes[j].coord -= diff;
          }
          left = node.coord + node.element.outerWidth() + gap;
        });

        // Reposition parent nodes to be centered relative to their children
        if (ll > 0) {
          layers[ll - 1].forEach(node => {
            if (node.children && node.children.length > 0) {
              var left = node.children[0].coord;
              var last = node.children[node.children.length - 1];
              var right = last.coord + last.element.outerWidth();
              node.coord = left + (right - left - node.element.outerWidth()) / 2;
            }
          });
        }
      }

      var newLayer: RenderedDerivationNode[] = [];
      layers[l].forEach(node => {
        if (node.phantom)
          return;

        if (!node.children || node.children.length == 0) {
          var phantomNode: RenderedDerivationNode = {
            sequent: node.sequent,
            element: node.element,
            parent: node,
            coord: node.coord,
            completed: node.completed,
            phantom: true
          };
          node.children = [phantomNode];
          newLayer.push(phantomNode);
          return;
        }

        var children = node.children;
        var nodeWidth = node.element.outerWidth();

        if (children.length == 1) {
          children[0].coord = node.coord - (children[0].element.outerWidth() - nodeWidth) / 2;
        }
        else if (children.length == 2) {
          children[0].coord = (node.coord + nodeWidth / 2 - gap / 2 - children[0].element.outerWidth());
          children[1].coord = (node.coord + nodeWidth / 2 + gap / 2);
        }

        newLayer.push.apply(newLayer, children);
      });

      if (newLayer.length > 0)
        layers.push(newLayer);
    }

    // Apply the calculated coordinates to the elements
    for (var l = 1; l < layers.length; ++l) {
      layers[l].forEach(node => {
        if (node.phantom)
          return;
        node.element.css('left', node.coord - node.parent.coord);
        leftMost = Math.min(leftMost, node.coord);
        rightMost = Math.max(rightMost, node.coord + node.element.outerWidth() + node.element.children('.rule').outerWidth());
      });
    }

    // Adjust the black lines and rule symbols
    layers.forEach(layer => {
      layer.forEach(node => {
        if (node.children && node.children.length > 0) {
          var rule = node.element.children('.rule');
          var symbolLeft = node.element.outerWidth();
          if (!node.children[0].phantom) {
            var lastChild = node.children[node.children.length - 1];
            var farLeft = node.children[0].coord;
            var farRight = lastChild.coord + lastChild.element.outerWidth();
            var diffLeft = farLeft - node.coord;
            var diffRight = node.coord + node.element.outerWidth() - farRight;
            if (diffLeft < 0)
              rule.css('margin-left', diffLeft);
            if (diffRight < 0) {
              rule.css('margin-right', diffRight);
              symbolLeft -= diffRight;
            }
          }
          rule.find('.symbol').css('left', symbolLeft);
        }
      });
    });

    return {right: rightMost, left: leftMost};
  }

  return {
    restrict: 'E',
    scope: {
      derivation: '=',
      selectedNode: '=',
      canInteract: '=',
      getLemmaLabel: '&'
    },
    link: (scope: DerivationDirectiveScope, element: JQuery, attrs: ng.IAttributes) => {
      var dragging = false;

      element.on('mousedown', (ev: JQueryMouseEventObject) => {
        if (!scope.canInteract)
          return;
        
        dragging = false;
        var origin = {x: ev.pageX, y: ev.pageY};
        var rootNode = element.children('.node');
        var rootNodeOrigin = rootNode.position();
        var moseMoveHandler = (ev: JQueryMouseEventObject) => {
          var deltaX = ev.pageX - origin.x;
          var deltaY = ev.pageY - origin.y;
          if (!dragging) {
            if (deltaX * deltaX + deltaY * deltaY < 16)
              return;
            dragging = true;
          }
          rootNode.css({
            top: rootNodeOrigin.top + deltaY,
            left: rootNodeOrigin.left + deltaX
          });
          ev.preventDefault();
        };
        $(document).on('mousemove', moseMoveHandler);
        $(document).one('mouseup', () => {
          $(document).off('mousemove', moseMoveHandler);
          $timeout(() => dragging = false, 0, false);
        });
      });

      element.on('click', '.node > .sequent', ev => {
        if (dragging || !scope.canInteract)
          return;

        var target = $(ev.target);
        element.find('.sequent.active').removeClass('active');
        target.addClass('active');
        scope.selectedNode = target.closest('.node').data('node');

        ev.stopPropagation();

        scope.$apply();
      });

      element.on('click', () => {
        if (dragging || !scope.canInteract)
          return;

        element.find('.sequent.active').removeClass('active');
        scope.selectedNode = null;
        scope.$apply();
      });

      scope.$watch('derivation', () => {
        var prevPos = element.children('.node').position();
        
        var rootNode = renderNode(scope.derivation, scope.selectedNode, taskId => scope.getLemmaLabel({lemmaTaskId: taskId}));
        element.empty().append(rootNode.element);
        layOutTree(rootNode);

        rootNode.element.css(prevPos || {
          top: element.innerHeight() - rootNode.element.outerHeight() - 10,
          left: (element.innerWidth() - rootNode.element.outerWidth()) / 2
        });
      }, true);
    }
  };
}
