interface SolutionResource extends ng.resource.IResource<SolutionResource> {
  completed: boolean;
  derivation: DerivationNode;
  steps: number;
  propositional: boolean;
}

interface SolutionResourceClass extends ng.resource.IResourceClass<SolutionResource> {
  getRules(): InferenceRule[];
}

SolutionResourceFactory.$inject = ['$resource', '$sce'];
function SolutionResourceFactory($resource: ng.resource.IResourceService, $sce: ng.ISCEService) {
  return $resource('$axiom/solutions/:taskId', {}, {
    get: {
      transformResponse: (data: string) => {
        if (!data) return null;
        var solution: SolutionResource = angular.fromJson(data);
        solution.derivation = angular.fromJson(<string><any>solution.derivation);
        DerivationNodes.setPaths(solution.derivation, '');
        return solution;
      }
    },
    getRules: {
      url: '$axiom/solutions/rules',
      isArray: true,
      transformResponse: (data: string) => {
        if (!data) return null;
        var rules: InferenceRule[] = angular.fromJson(data);
        rules.forEach((rule) => {
          rule.symbolHtml = $sce.trustAsHtml(InferenceRules.symbolToHtml(rule.symbol));
        });
        return rules;
      }
    }
  });
}

interface NodeReply {
  node: DerivationNode;
  steps: number;
  completed: boolean;
}

interface SolutionCtrlStateParams {
  taskId: number;
}

class SolutionCtrl {
  taskId: number;
  taskOrdinal: number;
  task: Task;
  
  rules: InferenceRule[];
  
  steps: number;
  completed: boolean;
  
  canInteract = false;
  
  selectedNode: DerivationNode;
  currentRule: InferenceRule;
  
  ruleInputPrompt: JQuery;
  
  inputPrompt: string;
  inputHint: any;
  input: string;
  ruleInputSubmitted = false;

  ws: WebSocket;
  wsDeferred: ng.IDeferred<NodeReply>;

  static $inject = ['$stateParams', '$modal', 'Solution', '$scope', '$location', 'auth', '$q', 'toast', 'taskset', 'translation', 'solution'];
  constructor($stateParams: SolutionCtrlStateParams,
              private $modal: ng.ui.bootstrap.IModalService,
              Solution: SolutionResourceClass,
              private $scope: ng.IScope,
              private $location: ng.ILocationService,
              private auth: AuthService,
              private $q: ng.IQService,
              private toast: ToastService,
              private taskset: TasksetResource,
              private translation: TranslationService,
              private solution: SolutionResource) {
    this.taskId = $stateParams.taskId;
    taskset.tasks.forEach((task: Task, i: number) => {
      if (task.id == this.taskId) {
        this.task = task;
        this.taskOrdinal = i + 1;
      }
    });
    
    this.rules = Solution.getRules();
    
    this.steps = solution.steps;
    this.completed = solution.completed;

    this.openWebSocket().then(() => {
      this.canInteract = true;
    });
    
    $scope.$on('$destroy', () => {
      if (this.ws && this.ws.readyState <= 1) {
        this.ws.close();
      }
    });
    
    this.ruleInputPrompt = $('.rule-input-prompt');
  }
  
  getLemmaLabel(lemmaTaskId: number): string {
    return (_.findIndex(this.taskset.tasks, {id: lemmaTaskId}) + 1).toString();
  }

  private openWebSocket(): ng.IPromise<void> {
    var deferredOpen = this.$q.defer<void>();
    
    this.ws = new WebSocket('ws://' + this.$location.host() + ':' + this.$location.port() + '/axiom/solve');
    this.ws.onopen = () => {
      this.ws.send(angular.toJson({
        token: this.auth.token,
        task: this.taskId
      }));
      deferredOpen.resolve();
    };
    
    this.ws.onmessage = (e) => {
      console.log(e.data);
      var data = angular.fromJson(e.data);
      if (data.prompt)
        this.wsDeferred.notify(data);
      else if (data.node)
        this.wsDeferred.resolve(data);
      else
        this.wsDeferred.reject(data);
    };
    
    this.ws.onerror = () => {
      deferredOpen.reject();
    };
    
    this.ws.onclose = () => {
      console.log('WebSocket closed');
    };
    
    return deferredOpen.promise;
  }

  applyRule(rule: InferenceRule) {
    if (!this.canInteract)
      return;

    var node = this.selectedNode;
    if (!node || node.rule || node.axiom || node.lemmaTaskId) {
      this.toast.error('solution.errors.SelectOpenNode');
      return;
    }
    
    console.log('Applying rule ' + rule.symbol);
    this.canInteract = false;
    this.currentRule = rule;
    this.wsDeferred = this.createDeferred();
    this.sendRuleApplication(node, rule);
  }
  
  private createDeferred() {
    var deferred = this.$q.defer<NodeReply>();
    deferred.promise.then((reply: NodeReply) => {
      this.steps = reply.steps;
      this.completed = reply.completed;
      this.updateSelectedNode(reply.node);
      this.finishRuleApplication();
    }, (error: any): any => {
      if (error && error.message)
        this.toast.error(error.message);
      
      if (this.ruleInputSubmitted) {
        this.ruleInputSubmitted = false;
        this.wsDeferred = this.createDeferred();
        return this.wsDeferred.promise;
      }
      this.finishRuleApplication();
    }, inputPrompt => {
      this.promptRuleInput(inputPrompt);
    });
    return deferred;
  }

  promptRuleInput(inputPrompt: {prompt: string; hint: any}) {
    this.ruleInputSubmitted = false;
    this.inputPrompt = inputPrompt.prompt;
    this.inputHint = inputPrompt.hint;
    this.input = inputPrompt.prompt == 'antecedentFormula' ? '0' : this.selectedNode.sequent;
    this.ruleInputPrompt.slideDown('fast', () => {
      this.ruleInputPrompt.find('input').select();
    });
  }

  onRuleInputKeyDown(ev: JQueryKeyEventObject) {
    if (ev.keyCode == 27) {
      ev.preventDefault();
      this.cancelRuleApplication();
    }
  }
  
  submitRuleInput() {
    if (this.input.toString().length > 0) {
      this.ruleInputSubmitted = true;
      this.sendRuleApplication(this.selectedNode, this.currentRule, this.input);
    }
  }
  
  cancelRuleApplication() {
    if (this.wsDeferred) {
      console.log('Cancelling rule application!');
      this.wsDeferred.reject();
    }
  }
  
  finishRuleApplication() {
    this.canInteract = true;
    this.currentRule = null;
    this.wsDeferred = null;
    this.ruleInputSubmitted = false;
    this.ruleInputPrompt.slideUp('fast');
    console.log('Finished rule application');
  }
  
  applyAxiom() {
    if (!this.canInteract)
      return;

    var node = this.selectedNode;
    if (!node || node.rule || node.axiom || node.lemmaTaskId) {
      this.toast.error('solution.errors.SelectOpenNode');
      return;
    }

    console.log('Marking as axiom');
    this.canInteract = false;
    this.wsDeferred = this.createDeferred();
    this.wsSend({
      action: 'axiom',
      nodePath: node.path
    });
  }
  
  selectLemma() {
    if (!this.canInteract)
      return;

    var node = this.selectedNode;
    if (!node || node.rule || node.axiom || node.lemmaTaskId) {
      this.toast.error('solution.errors.SelectOpenNode');
      return;
    }

    this.$modal.open({
      resolve: {
        numberedTasks: () => {
          var numberedTasks: {task: Task; ordinal: number}[] = [];
          for (var i = 0; i < this.taskset.tasks.length; ++i) {
            var task = this.taskset.tasks[i];
            if (task.id == this.taskId)
              break;
            numberedTasks.push({task: task, ordinal: i + 1});
          }
          return numberedTasks;
        }
      },
      templateUrl: 'templates/solution/selectLemma.html',
      controller: ['$scope', 'numberedTasks', function($scope: ng.IScope, numberedTasks: {task: Task; ordinal: number}[]) {
        $.extend($scope, {numberedTasks: numberedTasks});
      }]
    }).result.then((task: Task) => {
      this.canInteract = false;
      this.wsDeferred = this.createDeferred();
      this.wsSend({
        action: 'lemma',
        taskId: task.id,
        nodePath: node.path
      });
    });
  }
  
  clearNode() {
    var node = this.selectedNode;
    if (this.canInteract && node && (node.rule || node.axiom || node.lemmaTaskId)) {
      if (!confirm(this.translation.get('solution.ConfirmClearNode')))
        return;
      this.canInteract = false;
      this.wsDeferred = this.createDeferred();
      this.wsSend({
        action: 'clear',
        nodePath: node.path
      });
    }
  }
  
  private sendRuleApplication(node: DerivationNode, rule: InferenceRule, input: string = null) {
    this.wsSend({
      action: 'rule',
      rule: rule.symbol,
      nodePath: node.path,
      input: input
    });
  }
  
  private updateSelectedNode(newNode: DerivationNode) {
    var path = this.selectedNode.path;
    angular.copy(newNode, this.selectedNode);
    DerivationNodes.setPaths(this.selectedNode, path);
    if (this.selectedNode.children && this.selectedNode.children.length > 0 && this.currentRule)
      this.selectedNode = this.selectedNode.children[0];
  }
  
  private wsSend(data: any) {
    var promise: ng.IPromise<void> = <any>this.$q.when(null);
    if (this.ws.readyState > 1)
      promise = this.openWebSocket();
    promise.then(() => {
      this.ws.send(angular.toJson(data));
    });
  }
}
