interface Task {
  taskset?: TasksetResource;
  id?: number;
  sequent: string;
  steps?: number;
  text?: string;
  addText?: boolean;
  stepsDone?: number;
  completed?: boolean;
}

interface TasksetResource extends ng.resource.IResource<TasksetResource> {
  id?: number;
  name: string;
  open?: boolean;
  password?: string;
  needsPassword?: boolean;
  tasks?: Task[];
  taskCount?: number;
  lemmasAllowed?: boolean;
}

interface TasksetResourceClass extends ng.resource.IResourceClass<TasksetResource> {
  new(data: any): TasksetResource;
  getEditable(params: any): TasksetResource;
}

TasksetResourceFactory.$inject = ['$resource'];
function TasksetResourceFactory($resource: ng.resource.IResourceService) {
  return $resource('$axiom/tasksets/:id', {id: '@id'}, {
    getEditable: {url: '$axiom/tasksets/:id/edit'}
  });
}

interface TasksetIdStateParams {
  tasksetId: number;
}

class TasksetsCtrl {
  tasksets: TasksetResource[] = [];
  
  static $inject = ['currentUser', 'Taskset'];
  constructor(private currentUser: CurrentUserService,
              private Taskset: TasksetResourceClass) {
    this.tasksets = Taskset.query();
  }
  
  get canCreateTasksets(): boolean {
    return this.currentUser.isInRole(UserRole.INSTRUCTOR);
  }
  
  importTaskset() {
    var file = $('#taskset-import-file');
    file.on('change', () => {
      var files: FileList = file.prop('files');
      if (files.length == 0)
        return;
      var reader = new FileReader();
      reader.onload = () => {
        var taskset = JSON.parse(reader.result);
        this.Taskset.save(taskset, () => {
          window.location.reload();
        });
      };
      reader.readAsText(files[0]);
    });
    file.click();
  }
}

class TasksetCtrl {
  static $inject = ['$scope', 'Taskset', '$stateParams', '$state', 'taskset'];
  constructor($scope: ng.IScope,
              Taskset: TasksetResourceClass,
              $stateParams: TasksetIdStateParams,
              $state: ng.ui.IStateService,
              public taskset: TasksetResource) {
    
    $scope.$on('$stateChangeSuccess', (event: any, state: ng.ui.IState) => {
      if (state == $state.get('taskset.view'))
        Taskset.get({id: $stateParams.tasksetId}).$promise.then((newTaskset: TasksetResource) => {
          angular.copy(newTaskset, taskset);
        });
    });
  }
}

class TasksetViewCtrl {
  static $inject = ['$state', '$modal', 'currentUser', 'translation', 'taskset'];
  constructor(
      private $state: ng.ui.IStateService,
      private $modal: ng.ui.bootstrap.IModalService,
      private currentUser: CurrentUserService,
      private translation: TranslationService,
      private taskset: TasksetResource) {
  }
  
  get canViewSolutions(): boolean {
    return this.currentUser.isInRole(UserRole.ASSISTANT);
  }
  
  get canEdit() {
    return this.currentUser.isInRole(UserRole.INSTRUCTOR);
  }
  
  deleteTaskset() {
    if (confirm(this.translation.get('tasksets.ConfirmDeleteTaskset')))
      this.taskset.$delete(() => {
        this.$state.go('tasksets.list');
      });
  }
  
  exportTaskset() {
    this.$modal.open({
      resolve: {
        taskset: ['Taskset', (Taskset: TasksetResourceClass) => {
          return Taskset.getEditable({id: this.taskset.id}).$promise;
        }]
      },
      templateUrl: 'templates/tasksets/export.html',
      controller: ['$scope', 'taskset', function($scope: ng.IScope, taskset: TasksetResource) {
        var exported = {
          name: taskset.name,
          lemmasAllowed: taskset.lemmasAllowed,
          tasks: taskset.tasks.map(task => _.pick(task, 'sequent', 'steps', 'text'))
        };
        
        var blob = new Blob([JSON.stringify(exported)], {type: 'application/json'});
        var blobUrl = URL.createObjectURL(blob);
        $.extend($scope, {
          tasksetName: taskset.name,
          blobUrl: blobUrl
        });
        $scope.$on('$destroy', () => URL.revokeObjectURL(blobUrl));
      }]
    });
  }
}

class TasksetEditCtrl {
  editableTaskset: TasksetResource;
  
  static $inject = ['Taskset', '$state', 'toast', 'translation', 'taskset'];
  constructor(Taskset: TasksetResourceClass,
              private $state: ng.ui.IStateService,
              private toast: ToastService,
              private translation: TranslationService,
              private taskset: TasksetResource) {
    
    if (!this.taskset) {
      this.editableTaskset = new Taskset({tasks: [{}]});
    }
    else {
      this.editableTaskset = Taskset.getEditable({id: this.taskset.id});
      this.editableTaskset.$promise.then(() => {
        this.editableTaskset.tasks.forEach((task) => {
          task.addText = task.text != null;
        });
      });
    }
  }
  
  addTask() {
    this.editableTaskset.tasks.push({sequent: '', addText: false});
  }
  
  moveTaskUp(task: Task) {
    var i = this.editableTaskset.tasks.indexOf(task);
    this.editableTaskset.tasks.splice(i, 1);
    if (i == 0)
      this.editableTaskset.tasks.push(task);
    else
      this.editableTaskset.tasks.splice(i - 1, 0, task);
  }
  
  moveTaskDown(task: Task) {
    var i = this.editableTaskset.tasks.indexOf(task);
    this.editableTaskset.tasks.splice(i, 1);
    this.editableTaskset.tasks.splice(i == this.editableTaskset.tasks.length ? 0 : i + 1, 0, task);
  }
  
  removeTask(task: Task) {
    if (task.id && !confirm(this.translation.get('tasksets.ConfirmDeleteTask')))
      return;
    this.editableTaskset.tasks.splice(this.editableTaskset.tasks.indexOf(task), 1);
  }
  
  get canSubmit(): boolean {
    return (!this.taskset || this.editableTaskset.$resolved) && this.editableTaskset.tasks.length > 0;
  }
  
  saveTaskset() {
    this.editableTaskset.tasks.forEach((task) => {
      if (!task.addText)
        task.text = null;
    });
    this.editableTaskset.$save(() => {
      if (this.taskset)
        this.$state.go('taskset.view');
      else
        this.$state.go('tasksets.list');
    }, (response: {data: any[]}) => {
      response.data.forEach((error: {message: string; path: string; invalidValue: string}) => {
        var adder: ToastMessageAdder = null;
        if (error.path == 'Task.sequent') {
          this.editableTaskset.tasks.forEach((task, i) => {
            if (task.sequent == error.invalidValue)
              adder = this.toast.error('tasksets.errors.BadSequentInTask', [i + 1]);
          });
        }
        adder ? adder(error.message) : this.toast.error(error.message);
      });
    });
  }
}