enum UserRole {
  STUDENT, ASSISTANT, INSTRUCTOR, ADMIN
}

class CurrentUserService {
  uid: string;
  name: string;
  role: UserRole;
  
  private loadedDeferred: ng.IDeferred<{}>;
  loaded: ng.IPromise<{}>;
  
  private roleMap: {[role: string]: UserRole} = {
    STUDENT: UserRole.STUDENT,
    ASSISTANT: UserRole.ASSISTANT,
    INSTRUCTOR: UserRole.INSTRUCTOR,
    ADMIN: UserRole.ADMIN
  };
  
  static $inject = ['$http', '$q'];
  
  constructor(private $http: ng.IHttpService, $q: ng.IQService) {
    this.loadedDeferred = $q.defer();
    this.loaded = this.loadedDeferred.promise;
    
    $http.get('$axiom/auth').success((userInfo: UserResource) => {
      this.uid = userInfo.uid;
      this.name = userInfo.name;
      this.role = this.roleMap[userInfo.role];
      this.loadedDeferred.resolve();
    });
  }
  
  isPresent(): boolean {
    return this.uid != null;
  }
  
  isInRole(role: UserRole): boolean {
    return role <= this.role;
  }
  
  changePassword(newPassword: string): ng.IPromise<{}> {
    return this.$http.post('$axiom/auth/changepw', null, {params: {password: newPassword}});
  }
}

interface UserResource extends ng.resource.IResource<UserResource> {
  uid: string;
  name: string;
  sbn?: string;
  group?: string;
  password?: string;
  role: UserRole;
}

interface UserResourceClass extends ng.resource.IResourceClass<UserResource> {
  new(): UserResource;
  import(users: UserResource[]): ng.resource.IResource<{}>;
}

UserResourceFactory.$inject = ['$resource'];
function UserResourceFactory($resource: ng.resource.IResourceService) {
  var roleMap: {[role: string]: UserRole} = {
    STUDENT: UserRole.STUDENT,
    ASSISTANT: UserRole.ASSISTANT,
    INSTRUCTOR: UserRole.INSTRUCTOR,
    ADMIN: UserRole.ADMIN
  };
  
  return $resource('$axiom/users/:uid', {password: '@password'}, {
    query: {
      transformResponse: (data: string) => {
        if (!data) return null;
        var users: any[] = angular.fromJson(data);
        users.forEach((user: any) => {
          user.role = roleMap[user.role];
        });
        return users;
      },
      isArray: true
    },
    save: {
      method: 'POST',
      transformRequest: (user: UserResource) => {
        var encUser: any = user;
        encUser.role = UserRole[user.role];
        return angular.toJson(encUser);
      },
      transformResponse: (data: any) => {
        if (!data) return null;
        var user: any = angular.fromJson(data);
        user.role = roleMap[user.role];
        return user;
      },
      requestType: 'json' 
    },
    'import': {
      method: 'PUT',
      url: '$axiom/users/import'
    }
  });
}

class UsersCtrl {
  creatingNewUser = false;
  editableUser: UserResource;
  
  checked: {[uid: string]: boolean} = {};
  
  roles: {role: UserRole; label: string}[];
  
  groups: string[];
  selectedGroup: string;
  
  importUsersInput: string;
  
  static $inject = ['$state', 'User', 'currentUser', 'toast', 'translation', '$scope', 'users'];
  constructor(
      private $state: ng.ui.IStateService,
      private User: UserResourceClass,
      private currentUser: CurrentUserService,
      private toast: ToastService,
      private translation: TranslationService,
      $scope: ng.IScope,
      private users: UserResource[]) {
    
    var setRoles = () => {
      this.roles = [
        {role: UserRole.ADMIN, label: translation.get('users.roles.ADMIN')},
        {role: UserRole.INSTRUCTOR, label: translation.get('users.roles.INSTRUCTOR')},
        {role: UserRole.ASSISTANT, label: translation.get('users.roles.ASSISTANT')},
        {role: UserRole.STUDENT, label: translation.get('users.roles.STUDENT')}
      ].filter(r => r.role <= currentUser.role);
    };
    $scope.$on(TranslationService.EVT_LANG_CHANGE, setRoles);
    setRoles();
    
    $scope.$watch(() => this.users, () => {
      this.groups = _.uniq(this.users.map(u => u.group)).filter(g => !!g).sort();
    }, true);
  }
  
  canAddDeleteUsers() {
    return this.currentUser.isInRole(UserRole.INSTRUCTOR);
  }

  isCurrent(user: UserResource) {
    return user.uid == this.currentUser.uid;
  }
  
  canCheck(user: UserResource): boolean {
    return this.currentUser.uid != user.uid;
  }
  
  canDelete(user: UserResource): boolean {
    return this.canAddDeleteUsers() && !this.isCurrent(user);
  }
  
  roleLabel(role: UserRole): string {
    return this.translation.get('users.roles.' + UserRole[role]);
  }
  
  createNewUser() {
    this.creatingNewUser = true;
    this.editableUser = new this.User();
    this.editableUser.role = UserRole.STUDENT;
    angular.element('#modal-edit-user').modal('show');
  }
  
  editUser(user: UserResource) {
    this.creatingNewUser = false; 
    this.editableUser = angular.copy(user);
    angular.element('#modal-edit-user').modal('show');
  }
  
  canEditRole() {
    if (!this.canAddDeleteUsers())
      return false;
    return this.creatingNewUser || this.editableUser && !this.isCurrent(this.editableUser);
  }
  
  saveUser() {
    this.editableUser.$save({
      uid: this.creatingNewUser ? null : this.editableUser.uid
    }, (user: UserResource) => {
      _.remove(this.users, {uid: user.uid});
      this.users.push(user);
      angular.element('#modal-edit-user').modal('hide');
      this.toast.info(this.creatingNewUser ? 'users.UserCreated' : 'users.UserSaved', [user.name]);
    }, (response: {data: any[]}) => {
      this.toast.error(response.data[0].message);
    });
  }
  
  deleteUser(user: UserResource) {
    if (confirm(this.translation.get('users.ConfirmDeleteUser', [user.name])))
      user.$delete({uid: user.uid}, () => {
        _.remove(this.users, {uid: user.uid});
        this.toast.info('users.UserDeleted', [user.name]);
      });
  }
  
  importUsers() {
    var lines = this.importUsersInput.split('\n');
    var users: UserResource[] = [];
    lines.forEach(line => {
      var tokens = line.split('\t');
      var user = new this.User();
      user.sbn = tokens[0];
      user.name = tokens[1] + ' ' + tokens[2];
      // Either: [SBN] [Name] [Surname] [Group] (official format)
      // Or: [SBN] [Name] [Surname] [Junk] [Junk] [Group] (for easier importing from ÕIS)
      user.group = tokens.length > 4 ? tokens[5] : tokens[3];
      users.push(user);
    });
    
    this.User.import(users).$promise.then(() => {
      $('#modal-users-import').one('hidden.bs.modal', () => {
        this.$state.go(this.$state.current.name, {}, {reload: true});
      }).modal('hide');
    });
  }
}

UsersFilter.$inject = ['currentUser'];
function UsersFilter(currentUser: CurrentUserService) {
  return (users: UserResource[], group: string) => {
    if (!users) return users;
    var currentUid = currentUser.uid;
    if (group)
      users = users.filter(u => u.group == group || u.uid == currentUid);
    users.sort((u1, u2) => {
      if (u1.uid == currentUid) return -1;
      if (u2.uid == currentUid) return 1;
      if (u1.role != u2.role) return u2.role - u1.role;
      return u1.name.localeCompare(u2.name);
    });
    return users;
  };
}