package ee.ut.axiom;

import java.io.IOException;
import java.nio.file.Path;
import java.nio.file.Paths;
import javax.annotation.PreDestroy;
import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;
import javax.persistence.Persistence;

import org.glassfish.hk2.utilities.binding.AbstractBinder;
import org.glassfish.jersey.filter.LoggingFilter;
import org.glassfish.jersey.message.filtering.EntityFilteringFeature;
import org.glassfish.jersey.moxy.json.MoxyJsonConfig;
import org.glassfish.jersey.moxy.json.MoxyJsonFeature;
import org.glassfish.jersey.process.internal.RequestScoped;
import org.glassfish.jersey.server.ResourceConfig;
import org.glassfish.jersey.server.ServerProperties;
import org.glassfish.jersey.server.filter.RolesAllowedDynamicFeature;
import org.glassfish.jersey.server.validation.ValidationFeature;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.google.common.collect.ImmutableMap;

import ee.ut.axiom.models.User;
import ee.ut.axiom.repositories.SolutionRepository;
import ee.ut.axiom.repositories.TasksetRepository;
import ee.ut.axiom.repositories.UserRepository;

/**
 * @author Dmitri Gabbasov
 */
public class AxiomApplication extends ResourceConfig {
  private static Logger log = LoggerFactory.getLogger(AxiomApplication.class);

  private EntityManagerFactory emf;
  
  private Path homePath;

  public AxiomApplication(Path homePath) {
    this.homePath = homePath;
    
    emf = Persistence.createEntityManagerFactory("axiom",
        ImmutableMap.of("javax.persistence.jdbc.url",
            "jdbc:h2:" + homePath.resolve("data").resolve("axiom").toString() + ";AUTO_SERVER=TRUE"));

    registerComponents();
    
    createFirstUser();
  }

  private void createFirstUser() {
    EntityManager em = emf.createEntityManager();
    try {
      em.getTransaction().begin();
      UserRepository users = new UserRepository(em);
      if (users.getAll().isEmpty()) {
        log.info("No users in the database, will now create an admin user with username/password - admin/admin. Make sure you change the default password right away.");
        
        User user = new User();
        user.uid = "admin";
        user.name = "Administrator";
        user.role = User.ADMIN;
        user.passwordHash = User.hashPassword("admin");
        users.save(user);
      }
      em.getTransaction().commit();
    }
    finally {
      if (em.getTransaction().isActive())
        em.getTransaction().rollback();
      em.close();
    }
  }

  private void registerComponents() {
    property(ServerProperties.BV_SEND_ERROR_IN_RESPONSE, true);
    
    packages("ee.ut.axiom");

    register(new LoggingFilter());

    register(RolesAllowedDynamicFeature.class);
    register(EntityFilteringFeature.class);
    register(MoxyJsonFeature.class);
    register(ValidationFeature.class);

    register(new MoxyJsonConfig().resolver());

    register(new AbstractBinder() {
      @Override
      protected void configure() {
        try {
          bind(emf).to(EntityManagerFactory.class);
          bindFactory(AxiomEntityManagerFactory.class).to(EntityManager.class).in(RequestScoped.class);

          bind(UserRepository.class).to(UserRepository.class);
          bind(TasksetRepository.class).to(TasksetRepository.class);
          bind(SolutionRepository.class).to(SolutionRepository.class);

          bind(new MessagesService(homePath.resolve("langs"))).to(MessagesService.class);
        }
        catch (IOException e) {
          throw new RuntimeException(e);
        }
      }
    });
  }

  @PreDestroy
  private void preDestroy() {
    log.info("Shutting down...");
    emf.close();
  }

  public EntityManagerFactory getEntityManagerFactory() {
    return emf;
  }
}
