package ee.ut.axiom;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.nio.file.Path;
import java.nio.file.WatchEvent;
import java.nio.file.WatchKey;
import java.nio.file.WatchService;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import org.jvnet.hk2.annotations.Service;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.yaml.snakeyaml.Yaml;
import com.google.common.base.Charsets;
import com.google.common.io.Files;
import com.google.common.io.PatternFilenameFilter;

import static java.nio.file.StandardWatchEventKinds.ENTRY_CREATE;
import static java.nio.file.StandardWatchEventKinds.ENTRY_DELETE;
import static java.nio.file.StandardWatchEventKinds.ENTRY_MODIFY;
import static java.nio.file.StandardWatchEventKinds.OVERFLOW;

/**
 * @author Dmitri Gabbasov
 */
@Service
public class MessagesService {
  private static Logger log = LoggerFactory.getLogger(MessagesService.class);
  
  private final Path langsPath;

  public static class Language {
    public String id;
    public String name;
    public Map<String, String> messages;

    private Language(String id, String name, Map<String, String> messages) {
      this.id = id;
      this.name = name;
      this.messages = messages;
    }
  }
  
  private Map<String, Language> languages = new ConcurrentHashMap<>();

  public MessagesService(Path langsPath) throws IOException {
    this.langsPath = langsPath;
    log.info("Using langs path: {}", langsPath);

    Thread thread = new Thread(this::watch, "MessageService language file watcher");
    thread.setDaemon(true);
    thread.start();
    
    loadAllLanguages();
  }
  
  private void watch() {
    try {
      WatchService watcher = langsPath.getFileSystem().newWatchService();
      langsPath.register(watcher, ENTRY_CREATE, ENTRY_MODIFY, ENTRY_DELETE);
      
      WatchKey key;
      do {
        key = watcher.take();
        
        for (WatchEvent<?> event : key.pollEvents()) {
          WatchEvent.Kind<?> kind = event.kind();
          if (kind == OVERFLOW)
            continue;
          
          Path path = langsPath.resolve((Path) event.context());
          String langId = Files.getNameWithoutExtension(path.toString());

          if (kind == ENTRY_CREATE) {
            log.info("Adding new language '{}'", langId);
            languages.put(langId, loadLanguage(path.toFile()));
          }
          else if (kind == ENTRY_MODIFY) {
            log.info("Reloading language '{}'", langId);
            languages.put(langId, loadLanguage(path.toFile()));
          }
          else if (kind == ENTRY_DELETE) {
            log.info("Deleting language '{}'", langId);
            languages.remove(langId);
          }
        }
      }
      while (key.reset());
    }
    catch (IOException e) {
      log.error("Unexpected exception", e);
    }
    catch (InterruptedException ignored) {}
  }

  public void loadAllLanguages() {
    languages.clear();

    File[] langFiles = langsPath.toFile().listFiles(new PatternFilenameFilter(".+\\.yaml"));

    for (File langFile : langFiles) {
      Language language = loadLanguage(langFile);
      languages.put(language.id, language);
    }
  }

  @SuppressWarnings("unchecked")
  private static Language loadLanguage(File langFile) {
    Map<String, String> messages = new HashMap<>();
    Yaml yaml = new Yaml();
    try (BufferedReader reader = Files.newReader(langFile, Charsets.UTF_8)) {
      collect(messages, "", (Map<String, Object>) yaml.loadAs(reader, Map.class));
    }
    catch (Exception e) {
      log.error("Error reading language form file " + langFile, e);
    }

    String langId = Files.getNameWithoutExtension(langFile.getName());
    String langName = messages.remove("$name");
    return new Language(langId, langName, messages);
  }

  @SuppressWarnings("unchecked")
  private static void collect(Map<String, String> collection, String partialKey, Map<String, Object> map) {
    map.forEach((key, value) -> {
      if (partialKey != null && !partialKey.isEmpty())
        key = partialKey + '.' + key;
      if (value instanceof Map)
        collect(collection, key, (Map<String, Object>) value);
      else {
        if (!(value instanceof String))
          log.warn("Locale value for key {} is not String but {}", key, value.getClass().getSimpleName());
        collection.put(key, value.toString());
      }
    });
  }
  
  public Map<String, Language> getLanguages() {
    return Collections.unmodifiableMap(languages);
  }

  public Map<String, String> get(String lang) {
    return languages.get(lang).messages;
  }
}
