package ee.ut.axiom;

import java.util.Arrays;
import java.util.List;
import java.util.function.Consumer;
import java.util.function.Function;
import java.util.stream.Collectors;
import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;
import javax.persistence.NoResultException;

import org.glassfish.grizzly.http.HttpRequestPacket;
import org.glassfish.grizzly.websockets.DataFrame;
import org.glassfish.grizzly.websockets.DefaultWebSocket;
import org.glassfish.grizzly.websockets.ProtocolHandler;
import org.glassfish.grizzly.websockets.WebSocketListener;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import com.google.common.base.CharMatcher;
import com.google.common.collect.ImmutableMap;
import com.google.gson.Gson;

import ee.ut.axiom.models.Solution;
import ee.ut.axiom.models.Task;
import ee.ut.axiom.models.User;
import ee.ut.axiom.repositories.SolutionRepository;
import ee.ut.axiom.repositories.TasksetRepository;
import ee.ut.axiom.repositories.UserRepository;
import ee.ut.axiom.sequentCalculus.DerivationNode;
import ee.ut.axiom.sequentCalculus.Formula;
import ee.ut.axiom.sequentCalculus.InferenceRule;
import ee.ut.axiom.sequentCalculus.InputRequiredException;
import ee.ut.axiom.sequentCalculus.Sequent;

/**
 * @author Dmitri Gabbasov
 */
public class SolvingWebSocket extends DefaultWebSocket {
  private static final Logger log = LoggerFactory.getLogger(SolvingWebSocket.class);
  private final Gson gson = new Gson();

  private final EntityManagerFactory emf;
  private String authToken;
  private int taskId;
  
  private DerivationNode derivation;
  private Sequent rootSequent;

  private static class InitMessage {
    public String token;
    public int task;
  }
  
  private static class ActionMessage {
    public String action;
    public String nodePath;
    public String rule;
    public String input;
    public int taskId;
  }
  
  private static class NodeReply {
    public DerivationNode node;
    public int steps;
    public boolean completed;
  }
  
  public SolvingWebSocket(EntityManagerFactory emf, ProtocolHandler protocolHandler, HttpRequestPacket request, WebSocketListener... listeners) {
    super(protocolHandler, request, listeners);
    this.emf = emf;
  }

  @Override
  public void onMessage(String data) {
    super.onMessage(data);
    log.debug("Message received: {}", data);
    
    if (authToken == null) {
      InitMessage message = gson.fromJson(data, InitMessage.class);
      authToken = message.token;
      taskId = message.task;
      
      doWithSolution(solution -> {
        derivation = gson.fromJson(solution.derivation, DerivationNode.class);
        rootSequent = Sequent.parse(derivation.sequent);
      });
      return;
    }
    
    ActionMessage message = gson.fromJson(data, ActionMessage.class);
    DerivationNode node = getNodeByPath(message.nodePath);
      
    switch (message.action) {
      case "rule":
        InferenceRule rule = InferenceRule.getBySymbol(message.rule);
        try {
          applyRule(node, rule, message.input);
        }
        catch (InputRequiredException e) {
          promptForInput(Sequent.parse(node.sequent), e.hint);
          return;
        }
        break;
      case "axiom":
        applyAxiom(node);
        break;
      case "lemma":
        applyLemma(node, message.taskId);
        break;
      case "clear":
        clearNode(node);
        break;
      default:
        log.error("Unknown message received: {}", data);
        return;
    }

    int steps = countSteps(derivation);
    boolean completed = isCompleted(derivation);

    doWithSolution(solution -> {
      solution.steps = steps;
      solution.completed = completed;
      solution.derivation = gson.toJson(derivation);
    });

    NodeReply reply = new NodeReply();
    reply.node = node;
    reply.steps = steps;
    reply.completed = completed;
    send(gson.toJson(reply));
  }

  private boolean isCompleted(DerivationNode node) {
    return node.axiom != null || node.lemmaTaskId != null || node.rule != null && Arrays.stream(node.children).allMatch(this::isCompleted);
  }

  private int countSteps(DerivationNode node) {
    if (node.rule != null)
      return 1 + Arrays.stream(node.children).collect(Collectors.summingInt(this::countSteps));
    return node.axiom == null && node.lemmaTaskId == null ? 0 : 1;
  }

  private void doInTransaction(Consumer<EntityManager> action) {
    doInTransaction((EntityManager em) -> {
      action.accept(em);
      return null;
    });
  }
  
  private <R> R doInTransaction(Function<EntityManager, R> action) {
    EntityManager em = emf.createEntityManager();
    try {
      em.getTransaction().begin();
      R ret = action.apply(em);
      em.getTransaction().commit();
      return ret;
    }
    finally {
      if (em.getTransaction().isActive())
        em.getTransaction().rollback();
      em.close();
    }
  }

  private void doWithSolution(Consumer<Solution> action) {
    doInTransaction((EntityManager em) -> {
      UserRepository users = new UserRepository(em);
      SolutionRepository solutions = new SolutionRepository(em);

      User user = users.findByToken(authToken).orElseThrow(IllegalStateException::new);

      Solution solution;
      try {
        solution = solutions.getSolution(taskId, user.uid);
      }
      catch (NoResultException e) {
        solution = solutions.create(taskId, user.uid);
      }

      action.accept(solution);
    });
  }

  private void applyRule(DerivationNode node, InferenceRule rule, String input) {
    Sequent sequent = Sequent.parse(node.sequent);
    
    if (sequent.isPropositional() && !sequent.isDerivable())
      throw new IllegalStateException("solution.errors.NonDerivableSequent");

    log.debug("Applying rule {} to node {}", rule, node);

    Sequent[] results = rule.apply(sequent, input);
    
    if (rootSequent.isPropositional())
      for (Sequent result : results)
        if (!result.isPropositional())
          throw new IllegalStateException("solution.errors.FirstOrderLogicNotAllowed");

    node.rule = rule.symbol;
    node.ruleInput = input;
    
    node.children = new DerivationNode[results.length];
    for (int i = 0; i < results.length; i++) {
      node.children[i] = new DerivationNode(results[i].toString());
    }
  }

  private void promptForInput(Sequent sequent, String hint) {
    String prompt;
    Object hintObject = hint;
    
    if (hint.charAt(0) == '#') {
      prompt = "antecedentFormula";
      List<String> formulas = sequent.antecedent.stream().map(Formula::toString).collect(Collectors.toList());
      formulas.remove(formulas.size() - 1);
      hintObject = formulas;
    }
    else if (CharMatcher.JAVA_LOWER_CASE.matches(hint.charAt(0)))
      prompt = "term";
    else
      prompt = "formula";
    
    send(gson.toJson(ImmutableMap.of("prompt", prompt, "hint", hintObject)));
  }

  private void applyAxiom(DerivationNode node) {
    Sequent sequent = Sequent.parse(node.sequent);

    if (sequent.isPropositional() && !sequent.isDerivable())
      throw new IllegalStateException("solution.errors.NonDerivableSequent");

    log.debug("Marking node {} as an axiom", node);

    node.axiom = sequent.detectAxiom();
    if (node.axiom == null)
      throw new IllegalStateException("solution.errors.NotAnAxiom");
  }

  private void applyLemma(DerivationNode node, int taskId) {
    doInTransaction((EntityManager em) -> {
      TasksetRepository tasks = new TasksetRepository(em);
      Task currentTask = tasks.findTask(this.taskId).get();
      
      if (!currentTask.taskset.lemmasAllowed)
        throw new IllegalStateException("Lemmas not allowed.");

      Sequent lemmaSequent = null;
      for (Task task : currentTask.taskset.tasks) {
        if (task.id == taskId) {
          lemmaSequent = Sequent.parse(task.sequent);
          break;
        }
        if (task.equals(currentTask))
          throw new IllegalStateException("solution.errors.IllegalLemmaTask");
      }

      if (lemmaSequent == null)
        throw new IllegalStateException("solution.errors.IllegalLemmaTask");
      
      Sequent sequent = Sequent.parse(node.sequent);
      if (!lemmaSequent.unify(sequent))
        throw new IllegalStateException("solution.errors.LemmaNotApplicable");
      
      node.lemmaTaskId = taskId;
    });
  }

  private void clearNode(DerivationNode node) {
    node.rule = null;
    node.ruleInput = null;
    node.axiom = null;
    node.lemmaTaskId = null;
    node.children = null;
  }

  private DerivationNode getNodeByPath(String nodePath) {
    DerivationNode node = derivation;
    for (char ch : nodePath.toCharArray()) {
      if (ch == '0')
        node = node.children[0];
      else if (ch == '1')
        node = node.children[1];
      else
        throw new IllegalArgumentException("Illegal node path.");
    }
    return node;
  }

  @Override
  public void onClose(DataFrame frame) {
    super.onClose(frame);
    log.debug("Socket closed.");
  }
}
