package ee.ut.axiom.repositories;

import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;
import javax.inject.Inject;
import javax.persistence.EntityManager;
import javax.persistence.NoResultException;

import org.eclipse.persistence.config.QueryHints;
import org.jvnet.hk2.annotations.Service;
import com.google.common.base.Strings;

import ee.ut.axiom.models.Solution;
import ee.ut.axiom.models.Task;
import ee.ut.axiom.models.Taskset;
import ee.ut.axiom.models.TasksetPlacement;
import ee.ut.axiom.views.SimpleTaskset;
import ee.ut.axiom.views.TasksetWithSolutionInfo;

import static com.google.common.base.Preconditions.checkArgument;
import static com.google.common.base.Preconditions.checkNotNull;

/**
 * @author Dmitri Gabbasov
 */
@Service
public class TasksetRepository {
  private EntityManager em;

  @Inject
  public TasksetRepository(EntityManager em) {
    this.em = em;
  }

  public List<SimpleTaskset> getOpenTasksets() {
    return em.createQuery("select t from Taskset t where t.open=true order by t.placement.position", Taskset.class)
        .getResultList().stream().map(SimpleTaskset::new).collect(Collectors.toList());    
  }
  
  public List<SimpleTaskset> getAllTasksets() {
    return em.createQuery("select t from Taskset t order by t.placement.position", Taskset.class)
        .getResultList().stream().map(SimpleTaskset::new).collect(Collectors.toList());
  }
  
  public Taskset getTasksetById(int id) {
    return em.find(Taskset.class, id);
  }
  
  public TasksetWithSolutionInfo getTasksetWithSolutions(int id, String userId) {
    Taskset taskset = em.createQuery("select t from Taskset t where t.id=:id", Taskset.class)
        .setParameter("id", id).setHint(QueryHints.BATCH, "t.tasks").getSingleResult();

    List<Solution> solutions =
        em.createQuery("select s from Solution s join s.task t join t.taskset ts where ts.id=:id and s.user.uid=:userId",
            Solution.class).setParameter("id", id).setParameter("userId", userId).getResultList();

    return new TasksetWithSolutionInfo(taskset, solutions);
  }
  
  public Optional<Task> findTask(int id) {
    return Optional.ofNullable(em.find(Task.class, id));
  }

  public Taskset save(Taskset taskset) {
    checkNotNull(taskset.tasks);
    checkArgument(!taskset.tasks.isEmpty());
    
    if (Strings.isNullOrEmpty(taskset.password))
      taskset.password = null;

    for (Task task : taskset.tasks) {
      Task existing = em.find(Task.class, task.id);
      if (existing != null) {
        task.sequent = existing.sequent;
        task.solutions = existing.solutions;
      }
    }

    try {
      taskset.placement = em.createQuery("select p from TasksetPlacement p where p.taskset=:taskset", TasksetPlacement.class)
          .setParameter("taskset", taskset).getSingleResult();
    }
    catch (NoResultException e) {
      taskset.placement = placeTasksetAtEnd(taskset);
    }
    
    return em.merge(taskset);
  }

  private TasksetPlacement placeTasksetAtEnd(Taskset taskset) {
    TasksetPlacement placement = new TasksetPlacement();
    placement.taskset = taskset;
    placement.position = em.createQuery("select max(p.position) from TasksetPlacement p", Integer.class).getSingleResult() + 1;
    return placement;
  }

  public void deleteTaskset(int tasksetId) {
    em.remove(em.find(Taskset.class, tasksetId));
  }
}
