package ee.ut.axiom.resources;

import java.util.List;
import java.util.Set;
import javax.annotation.security.RolesAllowed;
import javax.inject.Inject;
import javax.validation.ConstraintViolation;
import javax.validation.ConstraintViolationException;
import javax.validation.Valid;
import javax.validation.Validator;
import javax.ws.rs.DELETE;
import javax.ws.rs.ForbiddenException;
import javax.ws.rs.GET;
import javax.ws.rs.POST;
import javax.ws.rs.PUT;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.Produces;
import javax.ws.rs.QueryParam;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.SecurityContext;

import org.hibernate.validator.constraints.NotBlank;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ee.ut.axiom.constraints.NonExisting;
import ee.ut.axiom.models.User;
import ee.ut.axiom.repositories.UserRepository;

/**
 * @author Dmitri Gabbasov
 */
@Path("/users")
@Produces(MediaType.APPLICATION_JSON)
public class UserResource {
  private static Logger log = LoggerFactory.getLogger(UserResource.class);
  
  @Inject
  UserRepository users;
  
  @Inject
  SecurityContext secCon;
  
  @Inject
  Validator validator;
  
  @GET
  @RolesAllowed(User.ASSISTANT)
  public List<User> list() {
    List<User> users = this.users.getAll();
    users.removeIf(user -> !canCurrentUserSee(user));
    return users;
  }
  
  @POST
  @RolesAllowed(User.INSTRUCTOR)
  public User create(
      @Valid @NonExisting(message = "users.errors.UidInUse") User user,
      @NotBlank(message = "users.errors.EmptyPassword") @QueryParam("password") String password) {
    user.passwordHash = User.hashPassword(password);
    return users.save(user);
  }
  
  @POST
  @Path("{uid}")
  @RolesAllowed(User.ASSISTANT)
  public User save(User user, @QueryParam("password") String password) {
    User existing = users.getByUid(user.uid);
    
    if (!canCurrentUserSee(existing))
        throw new ForbiddenException();
    
    if (secCon.isUserInRole(User.INSTRUCTOR)) {
      existing.name = user.name;
      existing.sbn = user.sbn;
      if (!isCurrentUser(user))
        existing.role = user.role;
    }
    
    existing.group = user.group;
    
    if (password != null && !password.isEmpty()) {
      existing.passwordHash = User.hashPassword(password);
      existing.activeTokens.clear();
    }

    validateUser(existing);

    return existing;
  }

  @DELETE
  @Path("{uid}")
  @RolesAllowed(User.INSTRUCTOR)
  public void delete(@PathParam("uid") String uid) {
    User user = users.getByUid(uid);
    if (user.isInRole(User.ADMIN) && !secCon.isUserInRole(User.ADMIN))
      throw new ForbiddenException();
    
    users.deleteUser(user);
  }
  
  @PUT
  @Path("/import")
  @RolesAllowed(User.INSTRUCTOR)
  public void importUsers(List<User> userList) {
    userList.forEach(user -> {
      user.uid = user.sbn.toLowerCase();
      if (users.findByUid(user.uid).isPresent())
        return;
      user.passwordHash = User.hashPassword(user.sbn);
      user.role = User.STUDENT;
      validateUser(user);
      users.save(user);
    });
  }

  private void validateUser(User user) {
    Set<ConstraintViolation<User>> violations = validator.validate(user);
    if (!violations.isEmpty())
      throw new ConstraintViolationException(violations);
  }

  private boolean isCurrentUser(User user) {
    return secCon.getUserPrincipal().getName().equals(user.uid);
  }

  private boolean canCurrentUserSee(User user) {
    if (secCon.isUserInRole(User.ADMIN))
      return true;
    
    if (secCon.isUserInRole(User.INSTRUCTOR))
      return !user.isInRole(User.ADMIN);
    
    if (secCon.isUserInRole(User.ASSISTANT))
      return User.STUDENT.equals(user.role);

    return isCurrentUser(user);
  }
}
