package ee.ut.axiom.sequentCalculus;

import java.util.Arrays;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;

import com.google.common.base.CharMatcher;
import com.google.common.collect.HashMultimap;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.SetMultimap;
import com.google.common.collect.Sets;
import com.google.common.primitives.Chars;

/**
 * @author Dmitri Gabbasov
 */
public class Formulas {
  
  public static class Group implements Formula {
    
    public Formula formula;

    public Group(Formula formula) {
      this.formula = formula;
    }

    @Override
    public int priority() {
      return 100;
    }

    @Override
    public Set<Character> getVariables() {
      return formula.getVariables();
    }

    @Override
    public SetMultimap<Character, Integer> getPredicates() {
      return formula.getPredicates();
    }

    @Override
    public void renameVariable(char oldVar, char newVar) {
      formula.renameVariable(oldVar, newVar);
    }

    @Override
    public boolean evaluate(Map<Character, Boolean> propVarValues) {
      return formula.evaluate(propVarValues);
    }

    @Override
    public boolean isPropositional() {
      return formula.isPropositional();
    }

    @Override
    public Formula copy() {
      return group(formula.copy());
    }

    @Override
    public String toString() {
      return Tokens.LEFT_PAREN + formula.toString() + Tokens.RIGHT_PAREN;
    }

    @Override
    public boolean equals(Object other) {
      return other instanceof Group && ((Group) other).formula.equals(formula);
    }
  }
  
  public static class Variable implements Formula {
    public char name;

    public Variable(char name) {
      this.name = name;
    }

    @Override
    public String toString() {
      return String.valueOf(name);
    }

    @Override
    public int priority() {
      return 100;
    }

    @Override
    public Set<Character> getVariables() {
      return Collections.emptySet();
    }

    @Override
    public SetMultimap<Character, Integer> getPredicates() {
      HashMultimap<Character, Integer> map = HashMultimap.create();
      map.put(name, 0);
      return map;
    }

    @Override
    public void renameVariable(char oldVar, char newVar) {}

    @Override
    public boolean evaluate(Map<Character, Boolean> propVarValues) {
      return propVarValues.get(name);
    }

    @Override
    public boolean isPropositional() {
      return true;
    }

    @Override
    public Formula copy() {
      return new Variable(name);
    }

    @Override
    public boolean equals(Object other) {
      return other instanceof Variable && ((Variable) other).name == name;
    }
  }

  public static class Negation implements Formula {
    public Formula formula;

    public Negation(Formula formula) {
      this.formula = formula;
    }

    @Override
    public String toString() {
      return Tokens.NEGATION + parens(formula, priority(), false);
    }

    @Override
    public int priority() {
      return 90;
    }

    @Override
    public Set<Character> getVariables() {
      return formula.getVariables();
    }

    @Override
    public SetMultimap<Character, Integer> getPredicates() {
      return formula.getPredicates();
    }

    @Override
    public void renameVariable(char oldVar, char newVar) {
      formula.renameVariable(oldVar, newVar);
    }

    @Override
    public boolean evaluate(Map<Character, Boolean> propVarValues) {
      return !formula.evaluate(propVarValues);
    }

    @Override
    public boolean isPropositional() {
      return formula.isPropositional();
    }

    @Override
    public Formula copy() {
      return negate(formula.copy());
    }

    @Override
    public boolean equals(Object other) {
      return other instanceof Negation && ((Negation) other).formula.equals(formula);
    }
  }

  public static abstract class BinaryOperator implements Formula {
    public char operator;
    private int priority;
    public Formula left;
    public Formula right;

    public BinaryOperator(char operator, int priority, Formula left, Formula right) {
      this.operator = operator;
      this.priority = priority;
      this.left = left;
      this.right = right;
    }

    @Override
    public String toString() {
      return parens(left, priority, false) + operator + parens(right, priority, true);
    }

    @Override
    public int priority() {
      return priority;
    }

    @Override
    public Set<Character> getVariables() {
      return Sets.union(left.getVariables(), right.getVariables());
    }

    @Override
    public SetMultimap<Character, Integer> getPredicates() {
      SetMultimap<Character, Integer> map = left.getPredicates();
      map.putAll(right.getPredicates());
      return map;
    }

    @Override
    public void renameVariable(char oldVar, char newVar) {
      left.renameVariable(oldVar, newVar);
      right.renameVariable(oldVar, newVar);
    }

    @Override
    public boolean isPropositional() {
      return left.isPropositional() && right.isPropositional();
    }

    @Override
    public boolean equals(Object other) {
      if (!(other instanceof BinaryOperator))
        return false;
      BinaryOperator op = (BinaryOperator) other;
      return op.operator == operator && op.left.equals(left) && op.right.equals(right);
    }
  }
  
  public static class Conjunction extends BinaryOperator {
    public Conjunction(Formula left, Formula right) {
      super(Tokens.CONJUNCTION, 40, left, right);
    }

    @Override
    public boolean evaluate(Map<Character, Boolean> propVarValues) {
      return left.evaluate(propVarValues) && right.evaluate(propVarValues);
    }

    @Override
    public Formula copy() {
      return and(left.copy(), right.copy());
    }
  }
  
  public static class Disjunction extends BinaryOperator {
    public Disjunction(Formula left, Formula right) {
      super(Tokens.DISJUNCTION, 30, left, right);
    }

    @Override
    public boolean evaluate(Map<Character, Boolean> propVarValues) {
      return left.evaluate(propVarValues) || right.evaluate(propVarValues);
    }

    @Override
    public Formula copy() {
      return or(left.copy(), right.copy());
    }
  }
  
  public static class Implication extends BinaryOperator {
    public Implication(Formula left, Formula right) {
      super(Tokens.IMPLICATON, 20, left, right);
    }

    @Override
    public boolean evaluate(Map<Character, Boolean> propVarValues) {
      return !left.evaluate(propVarValues) || right.evaluate(propVarValues);
    }

    @Override
    public Formula copy() {
      return imply(left.copy(), right.copy());
    }
  }
  
  public static class BiConditional extends BinaryOperator {
    public BiConditional(Formula left, Formula right) {
      super(Tokens.BICONDITIONAL, 10, left, right);
    }

    @Override
    public boolean evaluate(Map<Character, Boolean> propVarValues) {
      return left.evaluate(propVarValues) == right.evaluate(propVarValues);
    }

    @Override
    public Formula copy() {
      return equiv(left.copy(), right.copy());
    }
  }

  public static class Predicate implements Formula {
    public char name;
    public char[] args;

    public Predicate(char name, char[] args) {
      this.name = name;
      this.args = args;
    }

    @Override
    public String toString() {
      if (Character.isUpperCase(name) || args.length != 2)
        return String.format("%s%s%s%s", name, Tokens.LEFT_PAREN, Chars.join(String.valueOf(Tokens.COMMA), args), Tokens.RIGHT_PAREN);
      return String.format("%s%s%s", args[0], name, args[1]);
    }

    @Override
    public int priority() {
      return CharMatcher.JAVA_UPPER_CASE.matches(name) ? 100 : 80;
    }

    @Override
    public Set<Character> getVariables() {
      return ImmutableSet.copyOf(Chars.asList(args));
    }

    @Override
    public SetMultimap<Character, Integer> getPredicates() {
      HashMultimap<Character, Integer> map = HashMultimap.create();
      map.put(name, args.length);
      return map;
    }

    @Override
    public void renameVariable(char oldVar, char newVar) {
      for (int i = 0; i < args.length; i++)
        if (args[i] == oldVar)
          args[i] = newVar;
    }

    @Override
    public boolean evaluate(Map<Character, Boolean> propVarValues) {
      throw new IllegalFirstOrderLogicOperation();
    }

    @Override
    public boolean isPropositional() {
      return false;
    }

    @Override
    public Formula copy() {
      return new Predicate(name, args.clone());
    }

    @Override
    public boolean equals(Object other) {
      if (!(other instanceof Predicate))
        return false;
      Predicate pred = (Predicate) other;
      return pred.name == name && Arrays.equals(pred.args, args);
    }
  }

  public static abstract class Quantifier implements Formula {
    private char name;
    public char variable;
    public Formula formula;

    public Quantifier(char name, char variable, Formula formula) {
      this.name = name;
      this.variable = variable;
      this.formula = formula;
    }

    @Override
    public String toString() {
      return String.format("%s%s%s", name, variable, parens(formula, priority(), false));
    }

    @Override
    public int priority() {
      return 90;
    }

    @Override
    public Set<Character> getVariables() {
      return Sets.difference(formula.getVariables(), ImmutableSet.of(variable));
    }

    @Override
    public SetMultimap<Character, Integer> getPredicates() {
      return formula.getPredicates();
    }

    @Override
    public void renameVariable(char oldVar, char newVar) {
      if (newVar == variable)
        throw new VariableRenameClashException();
      formula.renameVariable(oldVar, newVar);
    }
    
    public void renameBoundVariable(char newVar) {
      formula.renameVariable(variable, newVar);
      variable = newVar;
    }

    @Override
    public boolean evaluate(Map<Character, Boolean> propVarValues) {
      throw new IllegalFirstOrderLogicOperation();
    }

    @Override
    public boolean isPropositional() {
      return false;
    }

    @Override
    public boolean equals(Object other) {
      if (!(other instanceof Quantifier))
        return false;
      Quantifier quant = (Quantifier) other;
      return quant.name == name && quant.variable == variable && quant.formula.equals(formula);
    }
  }
  
  public static class UniversalQuantifier extends Quantifier {
    public UniversalQuantifier(char variable, Formula formula) {
      super(Tokens.UNIVERSALQ, variable, formula);
    }

    @Override
    public Formula copy() {
      return forAll(variable, formula.copy());
    }
  }
  
  public static class ExistentialQuantifier extends Quantifier {
    public ExistentialQuantifier(char variable, Formula formula) {
      super(Tokens.EXISTENTIALQ, variable, formula);
    }

    @Override
    public Formula copy() {
      return exists(variable, formula.copy());
    }
  }
  
  public static Formula parse(String formula) {
    return new Parser(formula).parseFormula();
  }
  
  public static Formula group(Formula formula) {
    return new Group(formula);
  }
  
  public static Negation negate(Formula formula) {
    return new Negation(formula);
  }
  
  public static Conjunction and(Formula left, Formula right) {
    return new Conjunction(left, right);
  }
  
  public static Disjunction or(Formula left, Formula right) {
    return new Disjunction(left, right);
  }
  
  public static Implication imply(Formula left, Formula right) {
    return new Implication(left, right);
  }
  
  public static BiConditional equiv(Formula left, Formula right) {
    return new BiConditional(left, right);
  }
  
  public static UniversalQuantifier forAll(char variable, Formula formula) {
    return new UniversalQuantifier(variable, formula);
  }
  
  public static ExistentialQuantifier exists(char variable, Formula formula) {
    return new ExistentialQuantifier(variable, formula);
  }

  public static boolean isTautology(Formula formula) {
    Map<Character, Boolean> vars = new LinkedHashMap<>();
    formula.getPredicates().keySet().forEach(name -> vars.put(name, null));
    
    if (vars.size() > 26)
      throw new IllegalStateException("Too many variables.");

    for (int val = (1 << vars.size()) - 1; val >= 0; --val) {
      int i = 0;
      for (Map.Entry<Character, Boolean> entry : vars.entrySet())
        entry.setValue((val & (1 << i++)) != 0);
      
      if (!formula.evaluate(vars))
        return false;
    }
    
    return true;
  }
  
  public static boolean unify(Formula pattern, Formula subject, Map<Character, String> subst) {
    if (subject instanceof Group) {
      subject = ((Group) subject).formula;
      if (pattern instanceof Group)
        pattern = ((Group) pattern).formula;
    }
    
    if (pattern instanceof Variable) {
      char varName = ((Variable) pattern).name;
      String subjectStr = subject.toString();
      if (subst.containsKey(varName))
        return subst.get(varName).equals(subjectStr);
      subst.put(varName, subjectStr);
      return true;
    }
    if (pattern instanceof Negation && subject instanceof Negation) {
      return unify(((Negation) pattern).formula, ((Negation) subject).formula, subst);
    }
    if (pattern instanceof BinaryOperator && subject instanceof BinaryOperator) {
      BinaryOperator patternOp = (BinaryOperator) pattern;
      BinaryOperator subjectOp = (BinaryOperator) subject;
      if (patternOp.operator != subjectOp.operator)
        return false;
      return unify(patternOp.left, subjectOp.left, subst) && unify(patternOp.right, subjectOp.right, subst);
    }
    if (pattern instanceof Quantifier && subject instanceof Quantifier) {
      Quantifier qPattern = (Quantifier) pattern;
      Quantifier qSubject = (Quantifier) subject;
      if (qPattern.name == qSubject.name) {
        String prev = subst.put(qPattern.variable, String.valueOf(qSubject.variable));
        boolean result = unify(qPattern.formula, qSubject.formula, subst);
        subst.remove(qPattern.variable);
        if (prev != null)
          subst.put(qPattern.variable, prev);
        return result;
      }
    }
    if (pattern instanceof Predicate && subject instanceof Predicate) {
      Predicate patternPred = (Predicate) pattern;
      Predicate subjectPred = (Predicate) subject;
      
      if (patternPred.args.length != subjectPred.args.length || patternPred.name != subjectPred.name)
        return false;
      
      for (int i = 0; i < patternPred.args.length; ++i) {
        char patternVar = patternPred.args[i];
        String subjectVar = String.valueOf(subjectPred.args[i]);
        if (subst.containsKey(patternVar) && !subst.get(patternVar).equals(subjectVar))
          return false;
        subst.put(patternVar, subjectVar);
      }
      return true;
    }
    return false;
  }

  private static String parens(Formula formula, int basePriority, boolean strict) {
    if (strict && formula.priority() > basePriority || !strict && formula.priority() >= basePriority)
      return formula.toString();
    return Tokens.LEFT_PAREN + formula.toString() + Tokens.RIGHT_PAREN;
  }
}
