package ee.ut.axiom.sequentCalculus;

import java.util.List;
import java.util.Set;
import java.util.function.BiFunction;
import java.util.stream.Collectors;

import static ee.ut.axiom.sequentCalculus.Formulas.*;

/**
 * @author Dmitri Gabbasov
 */
public enum InferenceRule {
  
  CONJUCTION_RIGHT("Γ⊢A;Γ⊢B", "Γ⊢A&B", "⊢&", (sequent, input) -> {
    Sequent[] result = {sequent.copy(), sequent.copy()};
    for (int i = 0; i <= 1; ++i) {
      BinaryOperator conjunction = assertConjunction(result[i].succedent);
      result[i].succedent = i == 0 ? conjunction.left : conjunction.right;
    }
    return result;
  }),
  
  CONJUCTION_LEFT("Γ,A,B⊢C", "Γ,A&B⊢C", "&⊢", (sequent, input) -> {
    Sequent result = sequent.copy();
    List<Formula> ant = result.antecedent;
    BinaryOperator conjunction = assertConjunction(ant.remove(ant.size() - 1));
    ant.add(conjunction.left);
    ant.add(conjunction.right);
    return new Sequent[]{result};
  }),
  
  DISJUNCTION_RIGHT_FIRST("Γ⊢A", "Γ⊢A∨B", "⊢∨_1", (sequent, input) -> {
    Sequent result = sequent.copy();
    result.succedent = assertDisjunction(result.succedent).left;
    return new Sequent[]{result};
  }),
  
  DISJUNCTION_RIGHT_SECOND("Γ⊢B", "Γ⊢A∨B", "⊢∨_2", (sequent, input) -> {
    Sequent result = sequent.copy();
    result.succedent = assertDisjunction(result.succedent).right;
    return new Sequent[]{result};
  }),
  
  DISJUNCTION_LEFT("Γ,A⊢C;Γ,B⊢C", "Γ,A∨B⊢C", "∨⊢", (sequent, input) -> {
    Sequent[] result = {sequent.copy(), sequent.copy()};
    for (int i = 0; i <= 1; ++i) {
      List<Formula> ant = result[i].antecedent;
      BinaryOperator disjunction = assertDisjunction(ant.remove(ant.size() - 1));
      ant.add(i == 0 ? disjunction.left : disjunction.right);
    }
    return result;
  }),
  
  IMPLICATION_RIGHT("Γ,A⊢B", "Γ⊢A→B", "⊢→", (sequent, input) -> {
    Sequent result = sequent.copy();
    BinaryOperator succedent = assertImplication(result.succedent);
    result.antecedent.add(succedent.left);
    result.succedent = succedent.right;
    return new Sequent[]{result};
  }),

  MODUS_PONENS("Γ⊢A;Γ⊢A→B", "Γ⊢B", "⊢/→", (sequent, input) -> {
    Formula inputFormula = parse(requireInput(input, "A"));

    Sequent result1 = sequent.copy();
    Formula oldSuccedent = result1.succedent;
    result1.succedent = inputFormula;

    Sequent result2 = result1.copy();
    result2.succedent = imply(result2.succedent, oldSuccedent);

    return new Sequent[]{result1, result2};
  }),
  
  NEGATION_RIGHT("Γ,A⊢B;Γ,A⊢¬B", "Γ⊢¬A", "⊢¬", (sequent, input) -> {
    Sequent result1 = sequent.copy();
    result1.antecedent.add(assertNegation(result1.succedent).formula);
    result1.succedent = parse(requireInput(input, "B"));
    
    Sequent result2 = result1.copy();
    result2.succedent = negate(result2.succedent);
    
    return new Sequent[]{result1, result2};
  }),
  
  ELIMINATE_NEGATION("Γ⊢¬¬A", "Γ⊢A", "⊢/¬", (sequent, input) -> {
    Sequent result = sequent.copy();
    result.succedent = negate(negate(result.succedent));
    return new Sequent[]{result};
  }),
  
  BICONDITIONAL_RIGHT("Γ,A⊢B;Γ,B⊢A", "Γ⊢A~B", "⊢~", (sequent, input) -> {
    Sequent[] result = {sequent.copy(), sequent.copy()};
    for (int i = 0; i <= 1; ++i) {
      BinaryOperator conjunction = assertBiConditional(result[i].succedent);
      result[i].antecedent.add(i == 0 ? conjunction.left : conjunction.right);
      result[i].succedent = i == 0 ? conjunction.right : conjunction.left;
    }
    return result;
  }),
  
  BICONDITIONAL_LEFT("Γ⊢A", "Γ,A~B⊢B", "~⊢", (sequent, input) -> {
    Sequent result = sequent.copy();
    BinaryOperator bicond = assertBiConditional(result.removeLastAntecedentFormula());
    if (!unwrapGroup(bicond.right).equals(result.succedent))
      throw new RuleNotApplicableException();
    result.succedent = bicond.left;
    return new Sequent[]{result};
  }),
  
  UNIVERSALQ_RIGHT("Γ⊢A(y)", "Γ⊢∀xA(x)", "⊢∀", (sequent, input) -> {
    Sequent result = sequent.copy();
    UniversalQuantifier quantifier = assertUniversalQuantifier(result.succedent);
    renameBoundVarToUnused(quantifier, result.getVariables());
    result.succedent = quantifier.formula;
    return new Sequent[]{result};
  }),
  
  UNIVERSALQ_LEFT("Γ,A(t)⊢B", "Γ,∀xA(x)⊢B", "∀⊢", (sequent, input) -> {
    Sequent result = sequent.copy();
    UniversalQuantifier quantifier = assertUniversalQuantifier(result.removeLastAntecedentFormula());
    quantifier.formula.renameVariable(quantifier.variable, requireInput(input, "t").charAt(0));
    result.antecedent.add(quantifier.formula);
    return new Sequent[]{result};
  }),
  
  EXISTENTIALQ_RIGHT("Γ⊢A(t)", "Γ⊢∃xA(x)", "⊢∃", (sequent, input) -> {
    Sequent result = sequent.copy();
    ExistentialQuantifier quantifier = assertExistentialQuantifier(result.succedent);
    quantifier.formula.renameVariable(quantifier.variable, requireInput(input, "t").charAt(0));
    result.succedent = quantifier.formula;
    return new Sequent[]{result};
  }),
  
  EXISTENTIALQ_LEFT("Γ,A(y)⊢B", "Γ,∃xA(x)⊢B", "∃⊢", (sequent, input) -> {
    Sequent result = sequent.copy();
    Set<Character> freeVars = result.getVariables();
    ExistentialQuantifier quantifier = assertExistentialQuantifier(result.removeLastAntecedentFormula());
    renameBoundVarToUnused(quantifier, freeVars);
    result.antecedent.add(quantifier.formula);
    return new Sequent[]{result};
  }),
  
  WEAKEN_RIGHT("Γ⊢A", "Γ,B⊢A", "S+", (sequent, input) -> {
    Sequent result = sequent.copy();
    if (result.removeLastAntecedentFormula() == null)
      throw new RuleNotApplicableException();
    return new Sequent[]{result};
  }),
  
  PERMUTE_RIGHT("Γ,Δ,A⊢B", "Γ,A,Δ⊢B", "S~", (sequent, input) -> {
    if (sequent.antecedent.size() < 2)
      throw new RuleNotApplicableException();
    Sequent result = sequent.copy();
    int i = sequent.antecedent.size() == 2 ? 0 : Integer.parseInt(requireInput(input, "#"));
    result.antecedent.add(result.antecedent.remove(i));
    return new Sequent[]{result};
  });

  public String top;
  public String bottom;
  public String symbol;

  private BiFunction<Sequent, String, Sequent[]> rule;
  
  InferenceRule(String top, String bottom, String symbol, BiFunction<Sequent, String, Sequent[]> rule) {
    this.top = top;
    this.bottom = bottom;
    this.symbol = symbol;
    this.rule = rule;
  }
  
  public static InferenceRule getBySymbol(String symbol) {
    for (InferenceRule rule : InferenceRule.values()) {
      if (rule.symbol.equals(symbol))
        return rule;
    }
    return null;
  }
  
  public Sequent[] apply(Sequent sequent, String input) {
    Sequent[] sequents = rule.apply(sequent, input);
    for (Sequent s : sequents) {
      s.antecedent = s.antecedent.stream().map(InferenceRule::unwrapGroup).collect(Collectors.toList());
      s.succedent = unwrapGroup(s.succedent);
    }
    return sequents;
  }

  private static Formula unwrapGroup(Formula formula) {
    return formula instanceof Group ? ((Group) formula).formula : formula;
  }

  private static Conjunction assertConjunction(Formula formula) {
    if (!(formula instanceof Conjunction))
      throw new RuleNotApplicableException();
    return (Conjunction) formula;
  }
  
  private static Disjunction assertDisjunction(Formula formula) {
    if (!(formula instanceof Disjunction))
      throw new RuleNotApplicableException();
    return (Disjunction) formula;
  }
  
  private static Implication assertImplication(Formula formula) {
    if (!(formula instanceof Implication))
      throw new RuleNotApplicableException();
    return (Implication) formula;
  }
  
  private static BiConditional assertBiConditional(Formula formula) {
    if (!(formula instanceof BiConditional))
      throw new RuleNotApplicableException();
    return (BiConditional) formula;
  }

  private static Negation assertNegation(Formula formula) {
    if (!(formula instanceof Negation))
      throw new RuleNotApplicableException();
    return (Negation) formula;
  }

  private static UniversalQuantifier assertUniversalQuantifier(Formula formula) {
    if (!(formula instanceof UniversalQuantifier))
      throw new RuleNotApplicableException();
    return (UniversalQuantifier) formula;
  }

  private static ExistentialQuantifier assertExistentialQuantifier(Formula formula) {
    if (!(formula instanceof ExistentialQuantifier))
      throw new RuleNotApplicableException();
    return (ExistentialQuantifier) formula;
  }

  private static void renameBoundVarToUnused(Quantifier quantifier, Set<Character> usedVars) {
    for (int i = 0; i < 26; ++i) {
      char var = (char) ((char) ((quantifier.variable - 'a') + i) % 26 + 'a');
      if (!usedVars.contains(var)) {
        try {
          quantifier.renameBoundVariable(var);
        }
        catch (VariableRenameClashException e) {
          continue;
        }
        return;
      }
    }
    throw new RuleApplicationConstraintException();
  }
  
  private static String requireInput(String input, String hint) {
    if (input == null || input.isEmpty())
      throw new InputRequiredException(hint);
    return input;
  }
}
