package ee.ut.axiom.sequentCalculus;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Set;

import com.google.common.base.CharMatcher;
import com.google.common.collect.HashMultimap;
import com.google.common.collect.SetMultimap;
import com.google.common.collect.Sets;
import com.google.common.primitives.Chars;

/**
 * @author Dmitri Gabbasov
 */
class Parser {
  private final char[] input;
  private int pos = 0;
  
  Parser(String input) {
    this.input = input.toCharArray();
  }
  
  public Sequent parseSequent() {
    Sequent result = new Sequent();
    if (!consume(Tokens.TURNSTILE)) {
      do {
        result.antecedent.add(parseNonGroup());
      }
      while (consume(Tokens.COMMA));
      expect(Tokens.TURNSTILE);
    }
    result.succedent = parseNonGroup();
    assertEof();
    
    ArrayList<Formula> formulas = new ArrayList<>(result.antecedent);
    formulas.add(result.succedent);
    verifyPredicateArities(formulas);
    formulas.forEach(this::verifyBoundVariablesDontClash);
    
    return result;
  }

  public Formula parseFormula() {
    Formula formula = parseNonGroup();
    assertEof();
    verifyPredicateArities(Collections.singletonList(formula));
    verifyBoundVariablesDontClash(formula);
    return formula;
  }
  
  private Formula parseNonGroup() {
    int pos = this.pos;
    Formula formula = parseBiConditional();
    if (formula instanceof Formulas.Group)
      throw new ExcessiveParenthesisException(pos);
    return formula;
  }

  private Formula parseBiConditional() {
    List<Formula> formulas = new ArrayList<>();
    do {
      formulas.add(parseImplication());
    }
    while (consume(Tokens.BICONDITIONAL));

    return formulas.stream().reduce(Formulas::equiv).get();
  }

  private Formula parseImplication() {
    List<Formula> formulas = new ArrayList<>();
    do {
      formulas.add(parseDisjunction());
    }
    while (consume(Tokens.IMPLICATON));

    return formulas.stream().reduce(Formulas::imply).get();
  }

  private Formula parseDisjunction() {
    List<Formula> formulas = new ArrayList<>();
    do {
      formulas.add(parseConjunction());
    }
    while (consume(Tokens.DISJUNCTION));

    return formulas.stream().reduce(Formulas::or).get();
  }

  private Formula parseConjunction() {
    List<Formula> formulas = new ArrayList<>();
    do {
      formulas.add(parseNonBinary(true));
    }
    while (consume(Tokens.CONJUNCTION));

    return formulas.stream().reduce(Formulas::and).get();
  }

  private Formula parseNonBinary(boolean allowBinaryPred) {
    if (consume(Tokens.NEGATION)) {
      return Formulas.negate(parseNonBinary(false));
    }
    if (consume(Tokens.UNIVERSALQ)) {
      return Formulas.forAll(next(CharMatcher.JAVA_LOWER_CASE), parseNonBinary(false));
    }
    if (consume(Tokens.EXISTENTIALQ)) {
      return Formulas.exists(next(CharMatcher.JAVA_LOWER_CASE), parseNonBinary(false));
    }
    if (consume(Tokens.LEFT_PAREN)) {
      Formula f = parseNonGroup();
      expect(Tokens.RIGHT_PAREN);
      return Formulas.group(f);
    }

    CharMatcher matcher = CharMatcher.JAVA_UPPER_CASE;
    if (allowBinaryPred)
      matcher = matcher.or(CharMatcher.JAVA_LOWER_CASE);
    char next = next(matcher);
    
    if (Character.isUpperCase(next)) {
      if (consume(Tokens.PRED_LEFT)) {
        List<Character> args = new ArrayList<>();
        do {
          args.add(next(CharMatcher.JAVA_LOWER_CASE));
        }
        while (consume(Tokens.PRED_COMMA));
        expect(Tokens.PRED_RIGHT);
        return new Formulas.Predicate(next, Chars.toArray(args));
      }

      return new Formulas.Variable(next);
    }
    
    expect('=');
    return new Formulas.Predicate('=', new char[]{next, next(CharMatcher.JAVA_LOWER_CASE)});
  }

  private boolean eof() {
    return pos >= input.length;
  }
  
  private char next(CharMatcher matcher) {
    if (eof())
      throw new ParseException(pos, null, null);

    if (!matcher.matches(input[pos]))
      throw new ParseException(pos, null, input[pos]);
    
    return input[pos++];
  }

  private void expect(char symbol) {
    try {
      next(CharMatcher.is(symbol));
    }
    catch (ParseException e) {
      e.expected = symbol;
      throw e;
    }
  }
  
  private boolean consume(char symbol) {
    if (eof() || input[pos] != symbol)
      return false;
    ++pos;
    return true;
  }

  private void assertEof() {
    if (!eof())
      throw new ParseException(pos, null, input[pos]);
  }
  
  private void verifyPredicateArities(Collection<Formula> formulas) {
    SetMultimap<Character, Integer> predicateArities = HashMultimap.create();
    formulas.forEach(formula -> predicateArities.putAll(formula.getPredicates()));
    predicateArities.asMap().forEach((name, arities) -> {
      if (arities.size() != 1)
        throw new PredicateArityMismatchException(name);
    });
  }
  
  private void verifyBoundVariablesDontClash(Formula formula) {
    Set<Character> freeVars = formula.getVariables();
    try {
      verifyBoundVariablesDontClash(formula, freeVars);
    }
    catch (BoundVariableClashException e) {
      e.free = freeVars.contains(e.varName);
      throw e;
    }
  }

  private void verifyBoundVariablesDontClash(Formula formula, Set<Character> vars) {
    if (formula instanceof Formulas.Group)
      verifyBoundVariablesDontClash(((Formulas.Group) formula).formula, vars);
    else if (formula instanceof Formulas.Negation)
      verifyBoundVariablesDontClash(((Formulas.Negation) formula).formula, vars);
    else if (formula instanceof Formulas.BinaryOperator) {
      verifyBoundVariablesDontClash(((Formulas.BinaryOperator) formula).left, vars);
      verifyBoundVariablesDontClash(((Formulas.BinaryOperator) formula).right, vars);
    }
    else if (formula instanceof Formulas.Quantifier) {
      Formulas.Quantifier quantifier = (Formulas.Quantifier) formula;
      if (vars.contains(quantifier.variable))
        throw new BoundVariableClashException(quantifier.variable);
      verifyBoundVariablesDontClash(quantifier.formula, Sets.union(vars, Collections.singleton(quantifier.variable)));
    }
  }
}
